/****************************************************************************
**
** Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the test suite of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** No Commercial Usage
** This file contains pre-release code and may not be distributed.
** You may use this file in accordance with the terms and conditions
** contained in the Technology Preview License Agreement accompanying
** this package.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** If you have questions regarding the use of this file, please contact
** Nokia at qt-info@nokia.com.
**
**
**
**
**
**
**
**
** $QT_END_LICENSE$
**
****************************************************************************/



#include <QtTest/QtTest>
#include <QtCore/qlocale.h>
#include <qaudiooutput.h>
#include <qaudiodeviceinfo.h>
#include <qaudio.h>
#include <qaudioformat.h>


class tst_QAudioOutput : public QObject
{
    Q_OBJECT
public:
    tst_QAudioOutput(QObject* parent=0) : QObject(parent) {}

private slots:
    void initTestCase();
    void settings();
    void buffers();
    void notifyInterval();
    void pullFile();
    void pushFile();

private:
    bool          available;
    QAudioFormat  format;
    QAudioOutput* audio;
};

void tst_QAudioOutput::initTestCase()
{
    format.setFrequency(8000);
    format.setChannels(1);
    format.setSampleSize(8);
    format.setCodec("audio/pcm");
    format.setByteOrder(QAudioFormat::LittleEndian);
    format.setSampleType(QAudioFormat::UnSignedInt);

    // Only perform tests if audio output device exists!
    QList<QAudioDeviceInfo> devices = QAudioDeviceInfo::availableDevices(QAudio::AudioOutput);
    if(devices.size() > 0)
        available = true;
    else {
        qWarning()<<"NOTE: no audio output device found, no test will be performed";
        available = false;
    }
    audio = new QAudioOutput(format, this);
}

void tst_QAudioOutput::settings()
{
    if(available) {
        // Confirm the setting we added in the init function.
        QAudioFormat f = audio->format();

        QVERIFY(format.channels() == f.channels());
        QVERIFY(format.frequency() == f.frequency());
        QVERIFY(format.sampleSize() == f.sampleSize());
        QVERIFY(format.codec() == f.codec());
        QVERIFY(format.byteOrder() == f.byteOrder());
        QVERIFY(format.sampleType() == f.sampleType());
    }
}

void tst_QAudioOutput::buffers()
{
    if(available) {
        // Should always have a buffer size greater than zero.
        int store = audio->bufferSize();
        audio->setBufferSize(4096);
        QVERIFY(audio->bufferSize() > 0);
        audio->setBufferSize(store);
        QVERIFY(audio->bufferSize() == store);
    }
}

void tst_QAudioOutput::notifyInterval()
{
    if(available) {
        QVERIFY(audio->notifyInterval() == 1000);   // Default

        audio->setNotifyInterval(500);
        QVERIFY(audio->notifyInterval() == 500);    // Custom

        audio->setNotifyInterval(1000);             // reset
    }
}

void tst_QAudioOutput::pullFile()
{
    if(available) {
        QFile file(SRCDIR "4.wav");
        QVERIFY(file.exists());
        file.open(QIODevice::ReadOnly);

        QSignalSpy readSignal(audio, SIGNAL(notify()));
        QSignalSpy stateSignal(audio, SIGNAL(stateChanged(QAudio::State)));
        audio->setNotifyInterval(100);

        // Always have default states, before start
        QVERIFY(audio->state() == QAudio::StoppedState);
        QVERIFY(audio->error() == QAudio::NoError);
        QVERIFY(audio->elapsedUSecs() == 0);

        audio->start(&file);
        QTest::qWait(20); // wait 20ms
        // Check state, bytesFree() and periodSize() are valid non-zero values.
        QVERIFY(audio->state() == QAudio::ActiveState);
        QVERIFY(audio->error() == QAudio::NoError);
        QVERIFY(audio->periodSize() > 0);
        QVERIFY(audio->elapsedUSecs() > 10000 && audio->elapsedUSecs() < 800000);
        QVERIFY(stateSignal.count() == 1); // State changed to QAudio::ActiveState

        // Wait until finished...
        QTestEventLoop::instance().enterLoop(1);
        QCOMPARE(audio->processedUSecs(), qint64(692250));

#ifdef Q_OS_WINCE
        // 4.wav is a little less than 700ms, so notify should fire 4 times on Wince!
        QVERIFY(readSignal.count() >= 4);
#else
        // 4.wav is a little less than 700ms, so notify should fire 6 times!
        QVERIFY(readSignal.count() >= 6);
#endif
        audio->stop();
        QTest::qWait(20); // wait 20ms
        QVERIFY(audio->state() == QAudio::StoppedState);
        QVERIFY(audio->elapsedUSecs() == 0);
        // Can only check to make sure we got at least 1 more signal, but can be more.
        QVERIFY(stateSignal.count() > 1);

        file.close();
    }
}

void tst_QAudioOutput::pushFile()
{
    if(available) {
        QFile file(SRCDIR "4.wav");
        QVERIFY(file.exists());
        file.open(QIODevice::ReadOnly);

        const qint64 fileSize = file.size();

        QIODevice* feed = audio->start();

        char* buffer = new char[fileSize];
        file.read(buffer, fileSize);

        qint64 counter=0;
        qint64 written=0;
        while(written < fileSize) {
            written+=feed->write(buffer+written,fileSize-written);
            QTest::qWait(20);
            counter++;
        }
        QTestEventLoop::instance().enterLoop(1);

        QVERIFY(written == fileSize);
        QVERIFY(audio->processedUSecs() == 692250);

        audio->stop();
        file.close();
        delete [] buffer;
        delete audio;
    }
}

QTEST_MAIN(tst_QAudioOutput)

#include "tst_qaudiooutput.moc"
